import { toPointer } from '@ui-schema/json-pointer/toPointer';
import { getValueType } from '@ui-schema/json-schema/Validator';
import { ERROR_CONST_MISMATCH, ERROR_ENUM_MISMATCH, ERROR_MULTIPLE_OF, ERROR_PATTERN, ERROR_WRONG_TYPE, validateConst, validateContains, validateEnum, validateMultipleOf, validatePattern, arrayValidator, objectValidator, oneOfValidator, validateTypes, ERROR_NOT_SET, validateMinMaxString, validateMinMaxNumber, validateMinMaxArray, validateMinMaxObject } from '@ui-schema/json-schema/Validators';
import { ValidatorOutput } from '@ui-schema/ui-schema/ValidatorOutput';
import { List, Map, Record } from 'immutable';
export const standardValidators = [{
  id: 'ref',
  validate: (schema, value, params) => {
    if (!schema.has('$ref')) return;
    const refSchema = params.resource?.findRef(schema.get('$ref'));
    const applied = [];
    if (refSchema) {
      const result = params.validate(refSchema.value(), value, {
        ...params,
        keywordLocation: [...params.keywordLocation, '$ref'],
        parentSchema: undefined
      });
      applied.push(refSchema.value());
      applied.push(...(result.applied || []));
    }
    return {
      applied: applied
    };
  }
}, {
  id: 'type',
  validate: (schema, value, params) => {
    const type = schema.get('type');
    if (!type) return;
    if (typeof value === 'undefined' && type && params.instanceLocation.length === 0) {
      params.output.addError({
        error: ERROR_WRONG_TYPE,
        keywordLocation: toPointer([...params.keywordLocation, 'type']),
        instanceLocation: toPointer(params.instanceLocation),
        context: {
          actual: 'undefined'
        }
      });
      return;
    }
    if (typeof type !== 'undefined' && !validateTypes(value, type)) {
      params.output.addError({
        error: ERROR_WRONG_TYPE,
        keywordLocation: toPointer([...params.keywordLocation, 'type']),
        instanceLocation: toPointer(params.instanceLocation),
        context: {
          actual: getValueType(value)
        }
      });
    }
  }
}, {
  validate: (schema, value, params) => {
    const not = schema.get('not');
    if (!not || typeof value === 'undefined') return;
    const tmpNot = params.validate(not, value, {
      ...params,
      keywordLocation: [...params.keywordLocation, 'not'],
      parentSchema: undefined,
      output: new ValidatorOutput(),
      context: {}
    });
    if (tmpNot.valid) {
      params.output.addError({
        error: 'not-is-valid',
        keywordLocation: toPointer([...params.keywordLocation, 'not']),
        instanceLocation: toPointer(params.instanceLocation)
      });
    }
  }
}, {
  types: ['string'],
  validate: (schema, value, params) => {
    if (!schema.has('pattern')) return;
    if (!validatePattern(value, schema.get('pattern'))) {
      params.output.addError({
        error: ERROR_PATTERN,
        keywordLocation: toPointer([...params.keywordLocation, 'pattern']),
        instanceLocation: toPointer(params.instanceLocation),
        context: {
          pattern: schema.get('pattern'),
          patternError: schema.get('patternError')
        }
      });
    }
  }
}, {
  validate: (schema, value, params) => {
    if (!schema.has('const')) return;
    if (!validateConst(schema.get('const'), value)) {
      params.output.addError({
        error: ERROR_CONST_MISMATCH,
        keywordLocation: toPointer([...params.keywordLocation, 'const']),
        instanceLocation: toPointer(params.instanceLocation),
        context: {
          const: schema.get('const')
        }
      });
    }
  }
}, {
  validate: (schema, value, params) => {
    if (!schema.has('enum')) return;
    if (!validateEnum(schema.get('enum'), value)) {
      params.output.addError({
        error: ERROR_ENUM_MISMATCH,
        keywordLocation: toPointer([...params.keywordLocation, 'enum']),
        instanceLocation: toPointer(params.instanceLocation),
        context: {
          enum: schema.get('enum')
        }
      });
    }
  }
}, {
  types: ['number'],
  validate: (schema, value, params) => {
    if (!schema.has('multipleOf')) return;
    if (!validateMultipleOf(schema.get('multipleOf'), value)) {
      params.output.addError({
        error: ERROR_MULTIPLE_OF,
        keywordLocation: toPointer([...params.keywordLocation, 'multipleOf']),
        instanceLocation: toPointer(params.instanceLocation),
        context: {
          multipleOf: schema.get('multipleOf')
        }
      });
    }
  }
}, {
  types: ['string'],
  validate: (schema, value, params) => {
    if (!['minLength', 'maxLength'].some(keyword => schema.has(keyword))) return;
    validateMinMaxString(schema, value, params.output);
  }
}, {
  types: ['number'],
  validate: (schema, value, params) => {
    if (!['minimum', 'exclusiveMinimum', 'maximum', 'exclusiveMaximum'].some(keyword => schema.has(keyword))) return;
    validateMinMaxNumber(schema, value, params.output);
  }
}, {
  types: ['array'],
  validate: (schema, value, params) => {
    if (!['minItems', 'maxItems'].some(keyword => schema.has(keyword))) return;
    validateMinMaxArray(schema, value, params.output);
  }
}, {
  types: ['object'],
  validate: (schema, value, params) => {
    if (!['minProperties', 'maxProperties'].some(keyword => schema.has(keyword))) return;
    validateMinMaxObject(schema, value, params.output);
  }
}, {
  types: ['array'],
  validate: (schema, value, params) => {
    if (!['contains', 'minContains', 'maxContains'].some(keyword => schema.has(keyword))) return;
    validateContains(schema, value, params);
  }
}, {
  id: 'required',
  validate: (schema, value, params) => {
    const requiredList = schema?.get('required');
    if (!requiredList || params.recursive && !value) return;
    if (Map.isMap(value) || Record.isRecord(value)) {
      requiredList.forEach(requiredKey => {
        if (!value.has(requiredKey)) {
          params.output.addError({
            error: ERROR_NOT_SET,
            keywordLocation: toPointer([...params.keywordLocation, 'required']),
            instanceLocation: toPointer(params.instanceLocation),
            context: {
              requiredKey: requiredKey
            }
          });
        }
      });
    } else if (typeof value === 'object' && !Array.isArray(value)) {
      requiredList.forEach(requiredKey => {
        if (!Object.hasOwnProperty.call(value, requiredKey)) {
          params.output.addError({
            error: ERROR_NOT_SET,
            keywordLocation: toPointer([...params.keywordLocation, 'required']),
            instanceLocation: toPointer(params.instanceLocation),
            context: {
              requiredKey: requiredKey
            }
          });
        }
      });
    } else if (typeof value === 'undefined' || value === null) {
      requiredList.forEach(requiredKey => {
        params.output.addError({
          error: ERROR_NOT_SET,
          keywordLocation: toPointer([...params.keywordLocation, 'required']),
          instanceLocation: toPointer(params.instanceLocation),
          context: {
            requiredKey: requiredKey
          }
        });
      });
    }
  }
}, arrayValidator, objectValidator, {
  id: 'unevaluatedProperties',
  types: ['object'],
  validate: (schema, value, params) => {
    if (schema.get('unevaluatedProperties') === false && Map.isMap(value) && value.size) {
      return {
        afterAll: result => {
          const keys = value.keys();
          const evaluatedProperties = result.context?.evaluatedProperties;
          if (!evaluatedProperties) {
            params.output.addError({
              error: 'unevaluated-property',
              context: {
                keys: Array.from(keys)
              }
            });
            return;
          }
          for (const key of keys) {
            if (!evaluatedProperties.has(key)) {
              params.output.addError({
                error: 'unevaluated-property',
                context: {
                  key: key
                }
              });
              break;
            }
          }
        }
      };
    }
  }
}, {
  id: 'if',
  validate: (schema, value, params) => {
    const keyIf = schema.get('if');
    if (!keyIf || typeof value === 'undefined') return undefined;
    const keyThen = schema.get('then');
    const keyElse = schema.get('else');
    const tmpIf = params.validate(keyIf, value, {
      ...params,
      keywordLocation: [...params.keywordLocation, 'if'],
      instanceLocation: [],
      instanceKey: undefined,
      parentSchema: undefined,
      recursive: true,
      output: new ValidatorOutput(),
      context: {}
    });
    if (tmpIf.valid) {
      if (keyThen) {
        const result = params.validate(keyThen, value, {
          ...params,
          keywordLocation: [...params.keywordLocation, 'then'],
          parentSchema: undefined
        });
        return {
          applied: [keyThen, ...(result.applied || [])]
        };
      }
    } else {
      if (keyElse) {
        const result = params.validate(keyElse, value, {
          ...params,
          keywordLocation: [...params.keywordLocation, 'else'],
          parentSchema: undefined
        });
        return {
          applied: [keyElse, ...(result.applied || [])]
        };
      }
    }
  }
}, oneOfValidator, {
  id: 'allOf',
  validate: (schema, value, params) => {
    const schemas = schema.get('allOf');
    if (!schemas) return undefined;
    const applied = [...schemas.toArray()];
    if (List.isList(schemas)) {
      for (const schema of schemas) {
        const result = params.validate(schema, value, {
          ...params,
          keywordLocation: [...params.keywordLocation, 'allOf'],
          parentSchema: undefined
        });
        applied.push(...(result.applied || []));
      }
    }
    return {
      applied: applied
    };
  }
}, {
  id: 'anyOf',
  validate: (schema, value, params) => {
    const schemas = schema.get('anyOf');
    if (!schemas) return undefined;
    let errorCount = 0;
    let appliedSchema = undefined;
    const output = new ValidatorOutput();
    if (List.isList(schemas)) {
      for (const schema of schemas) {
        const context = {};
        const result = params.validate(schema, value, {
          ...params,
          keywordLocation: [...params.keywordLocation, 'anyOf'],
          parentSchema: undefined,
          recursive: true,
          output: new ValidatorOutput(),
          context: context
        });
        if (result.valid) {
          errorCount = 0;
          appliedSchema = schema;
          break;
        } else {
          errorCount++;
          result.errors.forEach(err => output.addError(err));
        }
      }
    }
    if (errorCount) {
      output.errors.forEach(err => params.output.addError(err));
    }
    return {
      applied: [appliedSchema]
    };
  }
}, {
  id: 'dependent',
  validate: (schema, value, params) => {
    const dependencies = schema.get('dependencies');
    const dependentSchemas = schema.get('dependentSchemas');
    const dependentRequired = schema.get('dependentRequired');
    if (!Map.isMap(value)) return;
    if (!dependencies && !dependentSchemas && !dependentRequired) return;
    const applied = [];
    value.keySeq().forEach(key => {
      const key_dependencies = dependencies ? dependencies.get(key) : undefined;
      const key_dependentSchemas = dependentSchemas ? dependentSchemas.get(key) : Map.isMap(key_dependencies) ? key_dependencies : undefined;
      const key_dependentRequired = dependentRequired ? dependentRequired.get(key) : List.isList(key_dependencies) ? key_dependencies : undefined;
      if (typeof value.get(key) !== 'undefined') {
        if (key_dependentSchemas) {
          applied.push(key_dependentSchemas);
          const result = params.validate(key_dependentSchemas, value, {
            ...params,
            keywordLocation: [...params.keywordLocation, 'dependentRequired'],
            instanceLocation: [],
            instanceKey: undefined,
            parentSchema: undefined
          });
          applied.push(...(result.applied || []));
        }
        if (key_dependentRequired) {
          applied.push(Map({
            required: key_dependentRequired
          }));
          key_dependentRequired.forEach(req => {
            if (typeof value.get(req) === 'undefined') {
              params.output.addError({
                error: ERROR_NOT_SET,
                keywordLocation: toPointer([...params.keywordLocation, 'dependentRequired', key]),
                instanceLocation: toPointer(params.instanceLocation)
              });
            }
          });
        }
      }
    });
    return {
      applied: applied
    };
  }
}];