import { ValidatorOutput } from '@ui-schema/ui-schema/ValidatorOutput';
import { getValueType } from './getValueType.js';
function deduplicateByLast(array) {
  const map = new Map();
  for (const item of array) {
    map.set(item.id || Symbol(), item);
  }
  return Array.from(map.values());
}
export function createRegister(validators) {
  const register = deduplicateByLast(validators).reduce((register, validator) => {
    if (validator.types) {
      validator.types?.forEach(type => {
        if (!register.valueHandlers.has(type)) {
          register.valueHandlers.set(type, []);
        }
        register.valueHandlers.get(type).push(validator.validate);
      });
    } else {
      register.handlers.push(validator.validate);
    }
    return register;
  }, {
    valueHandlers: new Map(),
    handlers: [],
    toJSON() {
      return registerToJSON(register, validators);
    }
  });
  return register;
}
export function makeParams() {
  return {
    instanceLocation: [],
    keywordLocation: []
  };
}
export function Validator(validators) {
  const register = createRegister(validators);
  function validate(schema, value, params = makeParams()) {
    const scopedParams = {
      ...params,
      context: params.context || {},
      output: params.output || new ValidatorOutput(),
      validate: params.validate || validate
    };
    const applied = [];
    const afterAll = [];
    for (const validator of register.handlers) {
      const result = validator(schema, value, scopedParams);
      applied.push(...(result?.applied || []));
      if (result?.afterAll) {
        afterAll.push(result.afterAll);
      }
    }
    const valueType = getValueType(value);
    if (valueType) {
      const validators = register.valueHandlers.get(valueType);
      if (validators) {
        for (const validator of validators) {
          const result = validator(schema, value, scopedParams);
          applied.push(...(result?.applied || []));
          if (result?.afterAll) {
            afterAll.push(result.afterAll);
          }
        }
      }
    }
    afterAll.forEach(afterAllValidator => {
      afterAllValidator({
        context: scopedParams.context,
        output: scopedParams.output
      });
    });
    if (scopedParams.output.errCount) {
      return {
        valid: false,
        errors: scopedParams.output.errors,
        applied: applied
      };
    }
    return {
      valid: true,
      value: value,
      applied: applied
    };
  }
  const validateValue = function (schema, value, params) {
    const output = new ValidatorOutput();
    validateValue.errors = output.errors;
    validateValue.annotations = output.annotations;
    return validate(schema, value, {
      instanceLocation: [],
      keywordLocation: [],
      ...(params || {}),
      output: output
    }).valid;
  };
  validateValue.errors = [];
  validateValue.annotations = [];
  return {
    validateValue: validateValue,
    validate: validate
  };
}
function mapToObject(map, valueConverter) {
  const obj = {};
  for (const [key, value] of map.entries()) {
    obj[String(key)] = valueConverter(value);
  }
  return obj;
}
function serializeValidatorsArray(validatorFns) {
  return function serializeValidatorsArray2(validators) {
    return validators.map(v => `[Validator Function ${validatorFns.indexOf(v)}]`);
  };
}
const registerToJSON = (register, validators) => {
  const validatorFns = validators.map(v => v.validate);
  return {
    valueHandlers: mapToObject(register.valueHandlers, serializeValidatorsArray(validatorFns)),
    handlers: serializeValidatorsArray(validatorFns)(register.handlers)
  };
};