import { isSomeSchema } from '@ui-schema/ui-schema/CommonTypings';
import { createOrdered } from '@ui-schema/ui-schema/createMap';
import { is, List, Map, OrderedMap, Set } from 'immutable';
const findMax = (currentVal, newVal) => {
  if (typeof newVal === 'number') {
    return typeof currentVal === 'number' ? Math.max(currentVal, newVal) : newVal;
  }
  return currentVal;
};
const findMin = (currentVal, newVal) => {
  if (typeof newVal === 'number') {
    return typeof currentVal === 'number' ? Math.min(currentVal, newVal) : newVal;
  }
  return currentVal;
};
const isFiniteNumber = n => typeof n === 'number' && Number.isFinite(n);
const getScale = n => {
  const s = n.toString().split('.')[1];
  return s ? 10 ** s.length : 1;
};
const getGCD = (a, b) => b === 0n ? a : getGCD(b, a % b);
const getLCM = (a, b) => a * b / getGCD(a, b);
export const findLeastCommonMultiple = (...values) => {
  if (!values.every(isFiniteNumber)) throw new Error('Non-finite input');
  if (values.includes(0)) return 0;
  const scaleFactors = values.map(getScale);
  const overallScale = scaleFactors.reduce((a, b) => a * b, 1);
  const intValues = values.map(v => BigInt(Math.round(v * overallScale)));
  return Number(intValues.reduce((a, b) => getLCM(a, b))) / overallScale;
};
export function mergeSchemas(baseSchema, ...appliedSchemas) {
  let mergedSchema = baseSchema instanceof Map ? baseSchema : OrderedMap(baseSchema);
  if (mergedSchema.has('allOf')) {
    mergedSchema = mergedSchema.remove('allOf');
  }
  let typeConflict = false;
  for (const appliedSchema of appliedSchemas) {
    if (!appliedSchema || !isSomeSchema(appliedSchema)) continue;
    let tmpSchema = mergedSchema;
    appliedSchema.forEach((value, key) => {
      switch (key) {
        case 'required':
          {
            const baseRequired = tmpSchema.get('required', List());
            const appliedRequired = value;
            tmpSchema = tmpSchema.set('required', baseRequired.toSet().union(appliedRequired.toSet()).toList());
            break;
          }
        case 'enum':
          {
            const baseEnum = tmpSchema.get('enum');
            const appliedEnum = value;
            if (baseEnum) {
              const baseSet = baseEnum.toSet();
              const appliedSet = appliedEnum.toSet();
              tmpSchema = tmpSchema.set('enum', baseSet.intersect(appliedSet).toList());
            } else {
              tmpSchema = tmpSchema.set('enum', appliedEnum);
            }
            break;
          }
        case 'type':
          {
            if (typeConflict) {
              break;
            }
            const baseType = tmpSchema.get('type');
            const appliedType = value;
            let baseTypes = Set(List.isList(baseType) ? baseType.toArray() : [baseType]);
            const appliedTypes = Set(List.isList(appliedType) ? appliedType.toArray() : [appliedType]);
            if (tmpSchema.has('type')) {
              baseTypes = baseTypes.intersect(appliedTypes);
            } else {
              baseTypes = appliedTypes;
            }
            if (baseTypes.size === 1) {
              tmpSchema = tmpSchema.set('type', baseTypes.first());
            } else if (baseTypes.size === 0) {
              typeConflict = true;
            } else {
              tmpSchema = tmpSchema.set('type', baseTypes.toList());
            }
            break;
          }
        case 'minLength':
        case 'minItems':
        case 'minimum':
        case 'exclusiveMinimum':
          tmpSchema = tmpSchema.set(key, findMax(tmpSchema.get(key), value));
          break;
        case 'maxLength':
        case 'maxItems':
        case 'maximum':
        case 'exclusiveMaximum':
          tmpSchema = tmpSchema.set(key, findMin(tmpSchema.get(key), value));
          break;
        case 'multipleOf':
          {
            const baseMultipleOf = tmpSchema.get('multipleOf');
            if (typeof baseMultipleOf === 'number' && typeof value === 'number') {
              tmpSchema = tmpSchema.set('multipleOf', findLeastCommonMultiple(baseMultipleOf, value));
            } else {
              tmpSchema = tmpSchema.set('multipleOf', value);
            }
            break;
          }
        case 'uniqueItems':
          tmpSchema = tmpSchema.set('uniqueItems', tmpSchema.get('uniqueItems', false) || value);
          break;
        case 'properties':
          {
            const baseProperties = tmpSchema.get('properties', OrderedMap());
            const appliedProperties = value;
            let newProperties = baseProperties;
            appliedProperties.forEach((subSchema, propKey) => {
              if (newProperties.has(propKey)) {
                newProperties = newProperties.updateIn([propKey, 'allOf'], (allOf = List()) => allOf.push(subSchema));
              } else {
                newProperties = newProperties.set(propKey, subSchema);
              }
            });
            tmpSchema = tmpSchema.set('properties', newProperties);
            break;
          }
        case 'patternProperties':
          {
            const basePatterns = tmpSchema.get('patternProperties', OrderedMap());
            const appliedPatterns = value;
            let newPatterns = basePatterns;
            appliedPatterns.forEach((subSchema, pattern) => {
              if (newPatterns.has(pattern)) {
                newPatterns = newPatterns.updateIn([pattern, 'allOf'], (allOf = List()) => allOf.push(subSchema));
              } else {
                newPatterns = newPatterns.set(pattern, subSchema);
              }
            });
            tmpSchema = tmpSchema.set('patternProperties', newPatterns);
            break;
          }
        case 'items':
        case 'prefixItems':
          {
            const isTuple = List.isList(value) && value.every(isSomeSchema);
            const base = tmpSchema.get(key);
            if (isTuple && List.isList(base)) {
              const merged = base.map((baseItem, idx) => {
                const appliedItem = value.get(idx);
                return appliedItem ? OrderedMap({
                  allOf: List([baseItem, appliedItem])
                }) : baseItem;
              });
              tmpSchema = tmpSchema.set(key, merged);
            } else {
              if (Map.isMap(value) && Map.isMap(base)) {
                tmpSchema = tmpSchema.set(key, OrderedMap({
                  allOf: List([base, value])
                }));
              } else {
                tmpSchema = tmpSchema.set(key, value);
              }
            }
            break;
          }
        case 'additionalProperties':
          if (value === false) {
            tmpSchema = tmpSchema.set('additionalProperties', false);
          } else if (Map.isMap(value) && tmpSchema.get('additionalProperties') !== false) {
            const baseAddProps = tmpSchema.get('additionalProperties');
            if (Map.isMap(baseAddProps)) {
              tmpSchema = tmpSchema.set('additionalProperties', OrderedMap({
                allOf: List([baseAddProps, value])
              }));
            } else {
              tmpSchema = tmpSchema.set('additionalProperties', value);
            }
          } else if (!tmpSchema.has('additionalProperties')) {
            tmpSchema = tmpSchema.set('additionalProperties', value);
          }
          break;
        case 'const':
          if (tmpSchema.has('const') && !is(tmpSchema.get('const'), value)) {
            tmpSchema = tmpSchema.update('allOf', (allOf = List()) => allOf.push(OrderedMap({
              const: value
            })));
          } else {
            tmpSchema = tmpSchema.set('const', value);
          }
          break;
        case 'allOf':
        case 'not':
        case 'if':
        case 'then':
        case 'else':
        case 'dependentSchemas':
        case '$defs':
        case 'definitions':
          break;
        case 'anyOf':
          {
            const baseAnyOf = tmpSchema.get('anyOf', List());
            const appliedAnyOf = key === 'anyOf' ? value : List();
            if (appliedAnyOf.size > 0) {
              if (baseAnyOf.size > 0) {
                tmpSchema = tmpSchema.set('anyOf', List([OrderedMap({
                  anyOf: baseAnyOf
                }), OrderedMap({
                  anyOf: appliedAnyOf
                })]));
              } else {
                tmpSchema = tmpSchema.set('anyOf', appliedAnyOf);
              }
            }
            break;
          }
        case 'oneOf':
          {
            const baseOneOf = tmpSchema.get('oneOf', List());
            const appliedOneOf = key === 'oneOf' ? value : List();
            if (appliedOneOf.size > 0) {
              if (baseOneOf.size > 0) {
                tmpSchema = tmpSchema.set('oneOf', List([OrderedMap({
                  oneOf: baseOneOf
                }), OrderedMap({
                  oneOf: appliedOneOf
                })]));
              } else {
                tmpSchema = tmpSchema.set('oneOf', appliedOneOf);
              }
            }
            break;
          }
        case '$ref':
          break;
        default:
          if (tmpSchema.has(key) && Map.isMap(tmpSchema.get(key)) && Map.isMap(value)) {
            tmpSchema = tmpSchema.set(key, tmpSchema.get(key).mergeDeep(value));
          } else {
            tmpSchema = tmpSchema.set(key, value);
          }
          break;
      }
    });
    mergedSchema = tmpSchema;
  }
  return mergedSchema;
}
export const validatorPlugin = {
  handle: props => {
    if (!props.validate) return {};
    const ownKey = props.storeKeys?.last();
    const result = props.validate(props.schema, props.value, {
      instanceLocation: props.storeKeys?.toArray() || [],
      keywordLocation: [],
      instanceKey: ownKey,
      parentSchema: props.parentSchema,
      resource: props.resource
    });
    const applicableSchema = mergeSchemas(props.schema, ...(result.applied || []));
    if (result.valid) {
      return {
        valid: true,
        schema: applicableSchema
      };
    }
    const errors = createOrdered(result.errors);
    return {
      valid: false,
      errors: errors,
      schema: applicableSchema
    };
  }
};