import { toPointer } from '@ui-schema/json-pointer/toPointer';
import { getValueType } from '@ui-schema/json-schema/Validator';
import { ValidatorOutput } from '@ui-schema/ui-schema/ValidatorOutput';
import { List } from 'immutable';
import { ERROR_WRONG_TYPE } from '@ui-schema/json-schema/Validators/TypeValidator';
export const ERROR_DUPLICATE_ITEMS = 'duplicate-items';
export const ERROR_NOT_FOUND_CONTAINS = 'not-found-contains';
export const ERROR_MIN_CONTAINS = 'min-contains';
export const ERROR_MAX_CONTAINS = 'max-contains';
export const ERROR_ADDITIONAL_ITEMS = 'additional-items';
const findDuplicates = arr => arr.filter((item, index) => arr.indexOf(item) !== index);
export const validateArrayContent = (itemsSchema, value, additionalItems = true, params) => {
  const output = 'output' in params && params.output ? params.output : new ValidatorOutput();
  let found = 0;
  if (List.isList(itemsSchema)) {
    if (List.isList(value) || Array.isArray(value)) {
      if (!validateAdditionalItems(additionalItems, value, itemsSchema)) {
        output.addError({
          error: ERROR_ADDITIONAL_ITEMS,
          keywordLocation: toPointer([...params.keywordLocation, 'additionalItems']),
          instanceLocation: toPointer(params.instanceLocation)
        });
        found++;
      }
      if (params.recursive) {
        const listSize = itemsSchema.size || 0;
        let i = 0;
        for (const val of value) {
          if (i >= listSize) break;
          const result = params.validate(itemsSchema.get(i), val, {
            ...params,
            keywordLocation: [...params.keywordLocation, 'items'],
            instanceLocation: [...params.instanceLocation, i],
            instanceKey: i,
            output: 'output' in params && params.output ? params.output : new ValidatorOutput(),
            context: {}
          });
          if (result.valid) {
            found++;
          } else if (!('output' in params && params.output)) {
            result.errors.forEach(err2 => output.addError(err2));
          }
          i++;
        }
      }
    } else {
      output.addError({
        error: ERROR_WRONG_TYPE,
        context: {
          actual: getValueType(value),
          arrayTupleValidation: true
        }
      });
      found++;
    }
  } else if (Array.isArray(value) || List.isList(value)) {
    let i = 0;
    for (const val of value) {
      const result = params.validate(itemsSchema, val, {
        ...params,
        keywordLocation: [...params.keywordLocation, 'items'],
        instanceLocation: [...params.instanceLocation, i],
        instanceKey: i,
        output: 'output' in params && params.output ? params.output : new ValidatorOutput(),
        context: {}
      });
      if (result.valid) {
        found++;
      } else if (!('output' in params && params.output)) {
        result.errors.forEach(err2 => output.addError(err2));
      }
      i++;
    }
  }
  return {
    output,
    found
  };
};
export const validateAdditionalItems = (additionalItems, value, schema) => {
  return additionalItems || !additionalItems && (List.isList(value) && value.size <= schema.size || Array.isArray(value) && value.length <= schema.size);
};
export const validateItems = (schema, value, params) => {
  const items = schema.get('items');
  if (items && value) {
    validateArrayContent(items, value, schema.get('additionalItems'), {
      ...params,
      parentSchema: schema
    });
  }
};
export const validateContains = (schema, value, params) => {
  if (!(Array.isArray(value) || List.isList(value))) return;
  const contains = schema.get('contains');
  if (!contains) return;
  const contains_type = contains.get('type');
  if (!contains_type) return;
  const minContains = schema.get('minContains');
  const maxContains = schema.get('maxContains');
  const item_err = validateArrayContent(contains, value, undefined, {
    ...params,
    parentSchema: schema,
    keywordLocation: [...params.keywordLocation, 'contains'],
    recursive: true,
    output: undefined,
    context: {}
  });
  if (item_err.found < 1 && typeof minContains === 'undefined' && typeof maxContains === 'undefined' || typeof minContains === 'number' && minContains > item_err.found || typeof maxContains === 'number' && maxContains < item_err.found) {
    if (item_err.output.errCount !== 0) {
      item_err.output.errors.forEach(err => params.output.addError(err));
    }
  }
  if (typeof minContains === 'number' && minContains > item_err.found) {
    params.output.addError({
      error: ERROR_MIN_CONTAINS,
      keywordLocation: toPointer([...params.keywordLocation, 'minContains']),
      instanceLocation: toPointer(params.instanceLocation),
      context: {
        minContains
      }
    });
  }
  if (typeof maxContains === 'number' && maxContains < item_err.found) {
    params.output.addError({
      error: ERROR_MAX_CONTAINS,
      keywordLocation: toPointer([...params.keywordLocation, 'maxContains']),
      instanceLocation: toPointer(params.instanceLocation),
      context: {
        maxContains
      }
    });
  }
  if (minContains !== 0 && (Array.isArray(value) && value.length === 0 || List.isList(value) && value.size === 0)) {
    params.output.addError({
      error: ERROR_NOT_FOUND_CONTAINS,
      keywordLocation: toPointer([...params.keywordLocation, 'contains']),
      instanceLocation: toPointer(params.instanceLocation)
    });
  }
};
export const validateUniqueItems = (schema, value) => {
  const uniqueItems = schema.get('uniqueItems');
  if (uniqueItems && (List.isList(value) || Array.isArray(value))) {
    const duplicates = findDuplicates(value);
    if (Array.isArray(duplicates)) {
      return duplicates.length === 0;
    } else if (List.isList(duplicates)) {
      return duplicates.size === 0;
    }
  }
  return true;
};
export const arrayValidator = {
  types: ['array'],
  validate: (schema, value, params) => {
    if (!schema || !value) return;
    const uniqueItems = schema?.get('uniqueItems');
    if (uniqueItems) {
      if (!validateUniqueItems(schema, value)) {
        params.output.addError({
          error: ERROR_DUPLICATE_ITEMS,
          keywordLocation: toPointer([...params.keywordLocation, 'uniqueItems']),
          instanceLocation: toPointer(params.instanceLocation)
        });
      }
    }
    const items = schema?.get('items');
    if (items && params.recursive) {
      validateItems(schema, value, params);
    }
    const contains = schema?.get('contains');
    if (contains) {
      validateContains(schema, value, params);
    }
  }
};